type bstmt =
    Lit of int    (* Push a literal *)
  | Drp           (* Discard a value *)
  | Bin of Ast.op (* Perform arithmetic on top of stack *)
  | Lfp of int    (* Load frame pointer relative *)
  | Sfp of int    (* Store frame pointer relative *)
  | Jsr of int    (* Call function by absolute address *)
  | Ent of int    (* Push FP, FP -> SP, SP += i *)
  | Rts of int    (* Restore FP, SP, consume formals, push result *)
  | Bra of int    (* Branch relative *)
  | Hlt           (* Terminate *)
  | StrConst of string
  | PStart of int * int * int
  | PEnd
  | InitInt of int
  | InitStr of int * int
  | InitArr of int * int
  | InitStructArr of int * int * int
  | Sstrfp of int * int
  | Lstrfp of int * int
  | Safp of int
  | Sasfp of int * int
  | Lafp of int
  | Lasfp of int * int
  | StctStart of int
  | StctEnd
  | FnStart of int
  | FnEnd
  | FnMain of int
  | ForStart
  | IfStart
  | ForEnd
  | StmtStart
  | StmtEnd
  | EOP
  | Semi
  | SemiIfNeed
  | ElseStart
  | Comma
  | Rparen
  |WhileStart

type prog = {
    num_globals : int;   (* Number of global variables *)
    text : bstmt array; (* Code for all the functions *)
  }

let string_of_stmt = function
    Lit(i) -> "Lit " ^ string_of_int i
  | Drp -> "Drp"
  | Bin(Ast.Add) -> "Add"
  | Bin(Ast.Sub) -> "Sub"
  | Bin(Ast.Mult) -> "Mul"
  | Bin(Ast.Div) -> "Div"
  | Bin(Ast.Equal) -> "Eql"
  | Bin(Ast.Neq) -> "Neq"
  | Bin(Ast.Less) -> "Lt"
  | Bin(Ast.Leq) -> "Leq"
  | Bin(Ast.Greater) -> "Gt"
  | Bin(Ast.Geq) -> "Geq"
  | Lfp(i) -> "Lfp " ^ string_of_int i
  | Sfp(i) -> "Sfp " ^ string_of_int i
  | Sstrfp(i,j) -> "Sstrfp " ^ string_of_int i ^ " " ^ string_of_int j 
  | Lstrfp(i,j) -> "Lstrfp " ^ string_of_int i ^ " " ^ string_of_int j 
  | Jsr(i) -> "Jsr " ^ string_of_int i
  | Ent(i) -> "Ent " ^ string_of_int i
  | Rts(i) -> "Rts " ^ string_of_int i
    | Bra(i) -> "Bra " ^ string_of_int i
  | PStart(i,j,k) -> "PStart" ^ string_of_int i ^ " " ^ string_of_int j ^ " " ^ string_of_int k
  | PEnd -> "PEnd" 
  | Hlt    -> "Hlt"
  | InitInt(i) -> "InitInt " ^ string_of_int i
  | InitStr(i,j) -> "InitStr " ^ string_of_int i ^ " " ^ string_of_int j
  | InitArr(i,j) -> "InitArr " ^ string_of_int i ^ " " ^ string_of_int j
  | InitStructArr(i,j,k) -> "InitStructArr " ^ string_of_int i ^ " " ^ string_of_int j ^ " " ^ string_of_int k
  | Safp(i) -> "Safp " ^ string_of_int i
  | Sasfp(i,j) -> "Safp " ^ string_of_int i ^ " " ^ string_of_int j
  | Lafp(i) -> "Lafp " ^ string_of_int i
  | Lasfp(i,j) -> "Lasfp " ^ string_of_int i ^ " " ^ string_of_int j
  | StctStart(i) -> "StctStart " ^ string_of_int i
  | StctEnd -> "StctEnd"
  | FnStart(i) -> "FnStart " ^ string_of_int i 
  | FnEnd -> "FnEnd" 
  | EOP -> "EOP"
  | FnMain (i) -> "FnMain " ^ string_of_int i
  | ForStart -> "ForStart"
  | ForEnd -> "ForEnd"
  | StmtStart -> "StmtStart"
  | StmtEnd -> "StmtEnd"
  | Semi -> "Semi"
  | SemiIfNeed -> "SemiIfNeed"
  | StrConst e -> "StrConst "  ^ e
  | IfStart -> "IfStart"
  | ElseStart -> "IfStart"
  | Comma -> "Comma"
  | Rparen -> "Rparen"
  | WhileStart -> "WhileStart"

  
let string_of_prog p =
  string_of_int p.num_globals ^ " global variables\n" ^
  let funca = Array.mapi
      (fun i s -> string_of_int i ^ " " ^ string_of_stmt s) p.text
  in String.concat "\n" (Array.to_list funca)
