open Ast
open Bytecode
open Semantic
open Sast
module StringMap = Map.Make(String)

(* Symbol table: Information about all the names in scope *)
type env = {
    function_index : int StringMap.t; (* Index for each function *)
    global_index   : int StringMap.t; (* "Address" for global variables *)
     struct_index   : struct_entry StringMap.t;
    local_index    : var_entry StringMap.t; (* FP offset for args, locals *)
  }

(* val enum : int -> 'a list -> (int * 'a) list *)
let rec enum stride n = function
    [] -> []
  | hd::tl -> (n, hd) :: enum stride (n+stride) tl

let rec build_local_idx map sidx offset = (function
    [] -> map
  | hd:: tl ->
    let offset1 = offset+1 in
     build_local_idx
    (StringMap.add hd.vname
        {
	  offset = offset1;
	  typ = hd.vtype;
        } map
      )
      sidx offset1 tl);;

let rec addSepList = (function
	[] -> []
	| hd::[] -> [hd]
	| hd::tl-> [hd; [Comma]] @ (addSepList tl));;
		
			

let rec addSep = (function

	[] -> []
	| hd::tl ->
		if (List.length tl > 0)
		then 
		let s = hd :: [Comma]
		in s @ addSep tl
		else [hd]
	)
;;


(*let calc_offset sidx offset typlst
=  offset + get_size_type sidx typlst;;*)
  

let rec build_local_idx_struct map sidx offset = (function
    [] -> map
  | hd:: tl ->
    let offset1 = offset+1 in
     build_local_idx
    (StringMap.add hd.vname
        {
	  offset = offset1;
	  typ = hd.vtype;
        } map
      )
      sidx offset1 tl);;




 let get_type_varname table varname =
    try (StringMap.find varname table).typ
    with Not_found -> raise (Failure("Type Checking Varname not found: " ^ varname))

 let get_offset_varname table varname =
    try (StringMap.find varname table).offset
    with Not_found -> raise (Failure("Type Checking Varname not found: " ^ varname))





(* val string_map_pairs StringMap 'a -> (int * 'a) list -> StringMap 'a *)
let string_map_pairs map pairs =
  List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs

(** Translate a program in AST form into a bytecode program.  Throw an
    exception if something is wrong, e.g., a reference to an unknown
    variable or function *)
let translate prog = 
let structs = prog.sdecls
  and globals = prog.gdecls
  	and functions = prog.fdecls  in

  (* Allocate "addresses" for each global variable *)
  let global_indexes = string_map_pairs StringMap.empty (enum 1 0 (List.map (fun f-> f.vname) globals)) in

  (* Assign indexes to function names; built-in "print" is special *)
  let built_in_functions = StringMap.add "print" (-1) StringMap.empty
   in let built_in_functions = StringMap.add "exit" (-2) built_in_functions
	in
  let function_indexes = string_map_pairs built_in_functions
      (enum 1 1 (List.map (fun f -> f.fname) functions)) in

  
let struct_indexes_t = 
	
	(enum 1 0 structs)
	in

  let struct_indexes = 
	
		List.fold_left
			  (fun map (index1,stct) ->
			    let soffset =  0 in
		      let mem_index = build_local_idx
				StringMap.empty map soffset (List.rev stct.smembers) in
				(
				  StringMap.add stct.sname
				    {
 				      index = index1;
				      memb_index = mem_index;
				    } map
				)
			  )
			  StringMap.empty struct_indexes_t
		
in
    (* Translate a function in AST form into a list of bytecode statements *)
  
let translate env fdecl =
    (* Bookkeeping: FP offsets for locals and arguments *)
    (*let local_offsets = enum 1 1 (List.map (fun f -> f.vname) fdecl.locals) in *)
    (*    let env = { env with local_index = string_map_pairs
		  StringMap.empty (local_offsets @ formal_offsets) }*)
		  let get_struct_table2 stct =
			    (try (StringMap.find stct env.struct_index).memb_index
			     with Not_found -> raise(Failure(" struct " ^ stct ^ " is not a type")))
			in
		  let get_struct_table typ_lst = 
		    match typ_lst with 
		    | [Struct(s)] -> (get_struct_table2 s)
			   | _ -> raise (Failure
		      ("Variable is not a Struct"))
		    in
	


		 let curr_offset = 0 in


			 let env =
			{
				env with local_index =
				(build_local_idx StringMap.empty env.struct_index curr_offset
				(fdecl.locals @ fdecl.formals))
			}


 in

   let child_offset base_offset base_typ table e1 =

	      let tab = (get_struct_table base_typ) in 

	      let child_offset = get_offset_varname tab e1
	
		in
		(base_offset,child_offset)


	

    in


    let get_base_c_offset e e1 table = 

		let base_offset = (StringMap.find e table).offset
			
		in
		let base_typ = get_type_varname table e in

		child_offset base_offset base_typ table e1
		
	in

 let get_arr_underlying_type array_id table =

		let localidxtyp = (StringMap.find array_id table).typ
		in

	 	List.hd (List.filter (fun (x) -> match x with | Arr(_) -> false
									| _ -> true) localidxtyp)

     in



    let rec expr table = function
	Literal i -> [Lit i]
	| String e ->	[StrConst e]	


      | Id s ->
	  (try [Lfp (StringMap.find s table).offset]
          with Not_found -> raise (Failure ("undeclared variable " ^ s)))
      | Binop (e1, op, e2) -> expr table e1  @ expr table e2  @ [Bin op]    

     
	(* Handling case of int i = a[0] *) 
	| Array(e,f) -> expr table f @
		( match e with
		| Id(t) -> 
			try [Lafp (StringMap.find t table).offset]
			  with Not_found -> raise (Failure ("undeclared variable " ^ t))
		| _ ->  raise (Failure ("Complex operations on arrays are disallowed ")))
		
	
	 
      | MemberAccess(ex,op,e1) -> 

			(	match ex
				with 
				(* Handling case of int i = a.x *)
				| Id e	->		
					let (b,c) = (get_base_c_offset e e1 table) 
					in
					[Lstrfp (b,c)]

				(* Handling case of int j = a[0].x *)
				| Array(e,f) ->	
					expr table f @
					(
					match e with
					| Id(t) ->
						let other_typ =  get_arr_underlying_type t table	
						in
						let b_offset =  (StringMap.find t table).offset in
						
						let (b,c) = child_offset b_offset [other_typ] table e1
				
						in						
						[Lasfp (b,c)]		

					
					| _ -> raise (Failure ("Complex operations on arrays are disallowed ")))


				
						
				|_ -> raise(Failure("Unexpected member access"))
		)
		 
      | Assign (s, e) ->
			let v1 = (expr table e)
			in
			(
			match s with
			(* Handling basic case of i = 1 *)
			| Id t -> v1@	
				(try [Sfp (StringMap.find t table).offset]
			  with Not_found -> raise (Failure ("undeclared variable " ^ t)))
				@ [SemiIfNeed]
			(* Handling case of a[0] = 1 *)
			| Array(e,f) -> v1 @ (expr table f @
				( match e with 
				| Id(t) -> 
				(try [Safp (StringMap.find t table).offset]
			  with Not_found -> raise (Failure ("undeclared variable " ^ t)))

				| _ -> raise (Failure ("Complex array assigments are disallowed "))))


			| MemberAccess(ex,op,e1) -> v1 @
						(
						match ex
						with
						(* handling case of a.x = 1  *) 
						| Id e	->		
							let (b,c) = (get_base_c_offset e e1 table) 
							in
							[Sstrfp (b,c)]

						(* handling case of a[0].x = 1  *)
						| Array(e,f) ->	expr table f @
							(
							match e with
							| Id(t) ->
							let other_typ =  get_arr_underlying_type t table	
							in
							let b_offset =  (StringMap.find t table).offset in
							
							let (b,c) = child_offset b_offset [other_typ] table e1
					
							in						
							[Sasfp (b,c)]		

							)
						| _ -> raise(Failure("Unexpected member acess")) 

					)
			)


	
      | Call (fname, actuals) -> 

	  let formals =  (List.map (expr table)  (actuals))

	in
	
	if (fname = "print")
	then

		List.concat formals @
	  [Jsr (StringMap.find fname env.function_index) ]   
	else
	(	
 
	  [Jsr (StringMap.find fname env.function_index) ]   
		@
	  	 List.concat (addSepList formals)
	  @ [Rparen] @ [SemiIfNeed]
	)
      	 
      | Noexpr -> []

    in let rec stmt table = function
      |	Block sl     ->  let mp = List.map (stmt table) sl
			in
		List.concat (mp)
      | Expr e       -> expr table e @ [SemiIfNeed]
      | If (p, t, f) -> let t' = stmt table t and f' = stmt table f in

	expr table p @ [IfStart] @ [StmtStart] @ t' @ [StmtEnd]
	@ [ElseStart] @ [StmtStart] @ f' @ [StmtEnd] 
	| While (cmp, stmt1) ->
		[WhileStart] @ (expr table cmp) @ [ForEnd] @ [StmtStart] @ (stmt table stmt1) @ [StmtEnd]	 
      | For (e1, e2, e3, b) ->

	  [ForStart] @ expr table e1 @ [SemiIfNeed] @ expr table e2 @ [SemiIfNeed] @ expr table e3
	 @ [ForEnd] 
	  @ [StmtStart] @ (stmt table b) @ [StmtEnd]
      | PMethod(a,b,c) ->
	
		let other_typ =  get_arr_underlying_type a table	
		in let offset_1 = (StringMap.cardinal env.local_index) + 1

		in

		let struct_index = match other_typ
			with 
			| Int -> -2
			|Struct(s) -> (StringMap.find s env.struct_index).index

		in
			
		
		let array_index = (StringMap.find a env.local_index).offset
		in
		
		let p_table = (StringMap.add b
				{
					offset = offset_1;
					typ = [other_typ];

				}
				StringMap.empty)
		in	
	
		let b' = Block c
		in	
		let c' = stmt p_table b'


		in
     
		[PStart (offset_1, struct_index, array_index)] @  c' @ [PEnd]	 


    in let initialize_generation local = match local.vtype with
	| [Arr(s);Int] ->
					
			(	match s with
				| Literal t -> [InitArr ((StringMap.find local.vname env.local_index).offset, t)]
						@ [SemiIfNeed]
					

				| _ -> [InitArr ((StringMap.find local.vname env.local_index).offset, 0)]
					@ [SemiIfNeed]
			)
		

	| [Arr(size);Struct(s)] ->
			let localidx = (StringMap.find local.vname env.local_index).offset
			in let structidx = StringMap.find s env.struct_index
			in

			(	match size with
				| Literal t -> [InitStructArr(localidx, structidx.index,t)]
					@ [SemiIfNeed]


				| _ -> [InitStructArr(localidx, structidx.index,0)]
					@ [SemiIfNeed]
			)


		(*	in InitStructArr(localidx, structidx.index,0)	*)
	| [Int] -> [InitInt (StringMap.find local.vname env.local_index).offset] 	@ [SemiIfNeed]

  	| [Struct(s)] ->
			let localidx = (StringMap.find local.vname env.local_index).offset
			in let structidx = StringMap.find s env.struct_index
			in
			[InitStr(localidx,structidx.index)]
			@ [SemiIfNeed]

	 in let initialize_generation_formal local = match local.vtype with
	| [Arr(s);Int] ->
					
			(	match s with
				| Literal t -> [InitArr ((StringMap.find local.vname env.local_index).offset, t)]
										

				| _ -> [InitArr ((StringMap.find local.vname env.local_index).offset, 0)]
			)
		

	| [Arr(size);Struct(s)] ->
			let localidx = (StringMap.find local.vname env.local_index).offset
			in let structidx = StringMap.find s env.struct_index
			in

			(	match size with
				| Literal t -> [InitStructArr(localidx, structidx.index,t)]

				| _ -> [InitStructArr(localidx, structidx.index,0)]
			)


		(*	in InitStructArr(localidx, structidx.index,0)	*)
	| [Int] -> [InitInt (StringMap.find local.vname env.local_index).offset]
  	| [Struct(s)] ->
			let localidx = (StringMap.find local.vname env.local_index).offset
			in let structidx = StringMap.find s env.struct_index
			in
			[InitStr(localidx,structidx.index)]
		

    

    in 
	let formals_1 = try addSep (List.concat (List.map initialize_generation_formal(fdecl.formals)))
	    with Not_found -> []
	in

	[FnStart (StringMap.find fdecl.fname function_indexes)] @      (* Entry: allocate space for locals *) 

	  formals_1
	@
	 [Rparen]
	@
	[StmtStart]
	@    
    List.concat (List.map initialize_generation (fdecl.locals)) @
    stmt env.local_index (Block fdecl.body) @  (* Body *)
    [FnEnd]   (* Default = return 0 *)

  in let initialize_struct_members key value acc = match value.typ with
     | [Int] -> acc@[(InitInt value.offset)]@[SemiIfNeed]
	

in

  let initialize_struct stct value acc =
	(
  	[StctStart value.index]
	@
	(StringMap.fold initialize_struct_members (value.memb_index) [])
	@
	[StctEnd])
	@ acc

in
(*
let initl_struct stc =
  	[StctStart (StringMap.find stc.sname struct_indexes).index]
	@
	List.map initialize_struct_members (stc.smembers)
	
	[StctEnd]
*)	

   
  (* Compile the functions *)
(*let struct_index_list = StringMap.to_list struct_indexes*)
	
  let struct_list = ((StringMap.fold initialize_struct struct_indexes []))


  in let env = { function_index = function_indexes;
		 global_index = global_indexes;
		 struct_index = struct_indexes;
		 local_index = StringMap.empty; } in

  (* Code executed to start the program: Jsr main; halt *)
  let entry_function = try
    [FnMain (StringMap.find "main" function_indexes); Hlt]
  with Not_found -> raise (Failure ("no \"main\" function"))
  in
   
  
  let func_bodies =  struct_list ::  entry_function :: List.map (translate env) functions
	 @[([EOP])] in

  

    { num_globals = List.length globals;
   text = Array.of_list(List.concat func_bodies)
  }
