open Ast
open Bytecode

   let outputCode code =
	let oc = open_out "faster_generated.c" in
	  print_endline code;
	  Printf.fprintf oc "%s" code;
	  close_out oc;;


let generate_prog prog =
  let code = Array.make 10 ""
  in 
  let stack = Array.make 1024 0
  in
  let variableCount = ref 0
  in
  let loopVariableCount = ref 0
	in
  let initial_code = "#include <stdio.h>\n#include <omp.h>\n#include <stdlib.h>"
	in
    let  globals = Array.make prog.num_globals 0 in

  let rec exec pc code = match prog.text.(pc) with
  | Lit i -> exec (pc+1) (string_of_int i :: code)
  | Drp -> exec (pc+1) code
  | Bin op -> let f::s::tail = code
	     in

	     let new_head = 
	     match op with
		Add -> s ^ " + " ^ f
		| Sub -> s ^ " - " ^ f
                | Mult -> s ^ " * " ^ f
		| Div -> s ^ " / " ^ f
		| Equal -> s ^ " == " ^ f
		| Neq -> s ^ " != " ^ f 
		| Less -> s ^ " < " ^ f
		| Leq -> s ^ " <= " ^ f
		| Greater -> s ^ " > " ^  f
		| Mod -> s ^ " % " ^  f

		| Geq -> s ^ " >= " ^ f
	    in exec (pc+1) (new_head::tail)
  | Sfp i ->  let h::t = code
		in			
		let stmt ="a_" ^ string_of_int i   ^ " = " ^  h
		in exec (pc+1) (stmt::t)
  | Lfp i -> let stmt = "a_" ^ string_of_int i;
		in exec (pc+1) (stmt::code)
  | Semi -> let stmt = ";\n"
		in exec (pc+1) (stmt::code)
  | StrConst e -> let stmt = e
		  in exec (pc+1) (stmt::code)

  | Jsr (-2) ->	let stmt = "exit(1\n"
		in exec (pc+1) (stmt::code)



  | Jsr (-1) -> let h::t = code in
		let stmt = 
		(if (String.contains h '"')
		then "printf(" ^ h ^ ");\n"  
		else "printf(\"%d\", " ^ h ^ ");\n" )
		in exec (pc+1) (stmt::t)

  | Jsr i -> 

		let stmt = "\nfunction_" ^ string_of_int i ^ "("
		 in exec (pc+1) (stmt::code)

  | Ent i -> exec (pc+1) code
  | Rts i -> print_endline (initial_code ^ (String.concat "" (List.rev code)))
  | Hlt ->   exec (pc+1) code
  | InitInt i ->variableCount :=  !variableCount + 1;
		let stmt = "int a_" ^ string_of_int i
	        in exec (pc+1) (stmt::code)  
  | InitArr (i,j) ->variableCount :=  !variableCount + 1;
		let stmt = "int a_" ^ string_of_int i ^ "[" ^ string_of_int j ^ "]"
	        in exec (pc+1) (stmt::code)
  | StctStart(i) -> let stmt = "\nstruct s_" ^ string_of_int i ^ "{\n"
		    in exec (pc+1) (stmt::code)
  | StctEnd -> let stmt = "};\n"
		in exec (pc + 1) (stmt::code)
  | InitStr (i,j) ->
		let stmt = "struct s_" ^ string_of_int j ^ " a_" ^ string_of_int i
		in exec (pc + 1) (stmt::code)
  | FnStart (i) ->
		let stmt = "\nfunction_" ^ string_of_int i ^ "(\n"
		in exec (pc + 1) (stmt::code)

  | FnEnd -> let stmt = "\n}\n" in  exec (pc + 1) (stmt::code)
  | EOP -> let output_code = initial_code ^ (String.concat "" (List.rev code))
	in outputCode output_code
	  

  | FnMain(i) -> let stmt = "\nmain(){\n function_" ^ string_of_int i ^ "();\n}"
		in exec (pc+1) (stmt::code)
  | ForStart -> let stmt = "\nfor("
		in exec (pc+1) (stmt::code)
  | ForEnd ->  let h::t = code
		in let stmt =   (try (String.sub h 0 (String.index h ';') ^ ")")
				with
				Not_found -> h ^ ")")
		in exec (pc+1) (stmt::t)

  | IfStart -> let h::t = code
		in let stmt = "\nif(" ^ h ^ ")"
		in exec (pc+1) (stmt::t)

  | ElseStart -> let h::t = code
		in let stmt = "\nelse"
		in exec (pc+1) (stmt::code)



   | StmtStart -> let stmt = "\n{\n"
		in exec (pc+1) (stmt::code)
		
   | StmtEnd -> let stmt = "\n}\n"
		in exec (pc+1) (stmt::code)
   | SemiIfNeed ->
		let h::t = code
		in let stmt =   (if (String.contains h ';') || (String.contains h ',')
				then "" else ";\n")
		in exec (pc+1) (stmt::code)

  | InitStructArr (i,j,k) ->
		let stmt = "struct " ^ "s_" ^ string_of_int j ^ " a_" ^ string_of_int i ^ "[" ^ string_of_int k ^ "]"
			in exec (pc+1) (stmt::code)

  | PStart (i,j,k)->
		(*	(local_index, struct_index, array_local_index() if integer struct_index = -2*)
		loopVariableCount :=  !loopVariableCount + 1;

		let loop_var_name = "loop_" ^ string_of_int !loopVariableCount
		in
		let int_loop_var_name = "int " ^ loop_var_name
		in
	
		let loop_var_decl = int_loop_var_name ^ " = " ^ "0;\n"
		in
		
		let pragma_call = "\n#pragma omp parallel for\n"
		in

		let size_length = "int length_" ^ string_of_int !loopVariableCount ^ " = " ^ "sizeof(a_" ^ string_of_int k ^ ")/sizeof(a_" ^ string_of_int k
			^ "[0]);\n" in 

	        let for_loop = "for(" ^  loop_var_name ^ " = " ^ "0; " ^ loop_var_name ^ " < " ^ "length_" ^ string_of_int !loopVariableCount ^ "; " ^ "++" ^ loop_var_name ^ "){\n" 	

		in
		let name_decl =
		if (j = -2)
		then "int a_" ^ string_of_int i ^ " = " ^ "a_" ^ string_of_int k ^ "[" ^ loop_var_name ^ "];\n"	
		else "struct s_" ^ string_of_int j ^ " a_" ^ string_of_int i ^ " = " ^ "a_" ^ string_of_int k ^ "[" ^ loop_var_name ^ "];\n"
		
		in let stmt = loop_var_decl ^ size_length ^  pragma_call ^ for_loop ^ name_decl
		in exec (pc+1) (stmt::code)  
   |Sstrfp(a,b) -> 
		let h::t = code
		in		
		let stmt = "a_" ^ string_of_int a ^ ".a_" ^ string_of_int b ^ " = " ^ h ^ ";\n"
		in exec (pc+1) (stmt::t)
   | Lstrfp(a,b) -> let stmt = "a_" ^ string_of_int a ^ "." ^ "a_" ^ string_of_int b
		in exec (pc+1) (stmt::code)


   | PEnd -> let stmt = "\n}\n"	in exec (pc+1) (stmt::code)

   | Lafp(i) ->
    		let h::t = code
		in			
		let stmt ="a_" ^ string_of_int i   ^ "[" ^  h ^ "]"
		in exec (pc+1) (stmt::t)
   | Safp(i) ->
    		let f::s::t = code
		in			
		let stmt ="a_" ^ string_of_int i   ^ "[" ^  f ^ "] = " ^ s ^ ";\n" 
		in exec (pc+1) (stmt::t)

   | Sasfp(i,j) ->
    		let f::s::t = code
		in			
		let stmt ="a_" ^ string_of_int i   ^ "[" ^  f ^ "].a_" ^ string_of_int j ^ " = " ^ s ^ ";\n" 
		in exec (pc+1) (stmt::t)

   | Lasfp(i,j) ->
    		let h::t = code
		in			
		let stmt ="a_" ^ string_of_int i   ^ "[" ^  h ^ "].a_" ^  string_of_int j
		in exec (pc+1) (stmt::t)

  | Comma -> let stmt = "," in exec (pc+1) (stmt::code)

  | Rparen -> let stmt = ")" in exec (pc+1) (stmt::code)
  | WhileStart -> let stmt = "\nwhile(" in exec (pc+1) (stmt::code)

	in
	exec 0 []

